/* CVS rsh client stuff.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.  */

#include <config.h>

#include "cvs.h"
#include "getline.h"
#include "buffer.h"

#ifdef CLIENT_SUPPORT

#include "rsh-client.h"

#ifndef NO_EXT_METHOD


/* TONY CHANGE */

/* Read cvsrc, processing options matching CMDNAME ("cvs" for global
   options, and update *ARGC and *ARGV accordingly.  */

#define	GROW	10

void
my_read_cvsrc(int *argc, char ***argv, const char *cmdname)
{
    FILE *cvsrcfile;

    char *line;
    int line_length;
    size_t line_chars_allocated;


    int command_len;
    int found = 0;

    int i;

    int new_argc;
    int max_new_argv;
    char **new_argv;

    cvsrcfile = open_cvsrc();
    if(cvsrcfile == NULL)
	return;

    /* now scan the file until we find the line for the command in question */

    line = NULL;
    line_chars_allocated = 0;
    command_len = strlen (cmdname);
    while ((line_length = getline (&line, &line_chars_allocated, cvsrcfile))
	   >= 0)
    {
	/* skip over comment lines */
	if (line[0] == '#')
	    continue;

	/* stop if we match the current command */
	if (!strncmp (line, cmdname, command_len)
	    && isspace ((unsigned char) *(line + command_len)))
	{
	    found = 1;
	    break;
	}
    }
    if (line_length < 0 && !feof (cvsrcfile))
	error (0, errno, "cannot read $HOME/.cvsrc");
    close_cvsrc();


    /* setup the options list */
    new_argc = 1;
    max_new_argv = GROW;
    new_argv =  (char **) xmalloc (max_new_argv * sizeof (char*));
    new_argv[0] = xstrdup(cmdname);

    if (found)
    {
        char *optstart;
	/* skip over command in the options line */
	for (optstart = strtok (line + command_len, "\t \n");
	     optstart;
	     optstart = strtok (NULL, "\t \n"))
	{
	    new_argv [new_argc++] = xstrdup (optstart);
	  
	    if (new_argc >= max_new_argv)
	    {
		max_new_argv += GROW;
		new_argv = (char **) xrealloc (new_argv, max_new_argv * sizeof (char*));
	    }
	}
    }

    if (line != NULL)
	free (line);

    *argc = new_argc;
    *argv = new_argv;

    return;
}
static const char *const user_cmd_usage[] =
{
    "Per user command are:\n",
    "        -r cmd      Set remote shell command\n",
    "        -rsh=cmd    Set remote shell command\n",
    "        -X arg      pass an extra argument to remote shell invocation\n",
    "                    May appear multiple time\n",
    NULL
};

/* A block of extra arguments to the remote shell command.
 * This is meant for options unique to a particular remote
 * shell command, and thus not supported directly by cvs */
static int	rsh_extra_argc = 0;
static int	max_rsh_extra_arg = 0;
static char	**rsh_extra_argv = NULL;


/* pick_rsh - gather per user@host specific arguments for the
 * remote shell command */

static char *pick_rsh(cvsroot_t *root)
{
    static const char short_options[] = "r:X:";
    static struct option long_options[] =
    {
        {"rsh", required_argument, NULL, 'r'},
        {"X", required_argument, NULL, 'X'},
        {0, 0, 0, 0}
    };
    /* `getopt_long' stores the option index here, but right now we
        don't use it. */
    int option_index = 0;

    char *ret = RSH_DFLT;

    if (root != NULL
	&& (root->username != NULL || root->hostname != NULL) )
    {
	int argc = 0;
	char **argv;
	char *junk_args[2] = { "", NULL };
	char *t_name;
	int alloc_len; 
	int c;

	alloc_len = 1;
	if(root->username != NULL) alloc_len += strlen(root->username);
	if(root->hostname != NULL) alloc_len += strlen(root->hostname) + 1;

	t_name = (char *) xmalloc (alloc_len);
	t_name[0] = '\0';
	if(root->username != NULL) strcat (t_name, root->username);
	if(root->hostname != NULL)
	{
	    strcat (t_name, "@");
	    strcat (t_name, root->hostname);
	}
	argv = junk_args;

	my_read_cvsrc(&argc, (char ***) &argv, t_name);

	optind = 0;
	opterr = 1;
	while ((c = getopt_long
            (argc, argv, short_options, long_options, &option_index))
           != EOF)
	{
	    switch (c)
	    {
		case 'r':
		    /* going to leak memory, but the client shuts down soon */
		    ret = strdup (optarg);
		    break;
		case 'X':
		    if (rsh_extra_argc >= max_rsh_extra_arg) {
			max_rsh_extra_arg += GROW;
			rsh_extra_argv = (char **) xrealloc (rsh_extra_argv, max_rsh_extra_arg * sizeof (char*));
		    }
		    rsh_extra_argv[rsh_extra_argc++] = strdup (optarg);
		    break;
	        default:
		    usage(user_cmd_usage);
	    }
	}
	free (t_name);
    }
    return ret;
}
/* TONY CHANGE */


/* Contact the server by starting it with rsh.  */

/* Right now, we have two different definitions for this function,
   depending on whether we start the rsh server using popenRW or not.
   This isn't ideal, and the best thing would probably be to change
   the OS/2 port to be more like the regular Unix client (i.e., by
   implementing piped_child)... but I'm doing something else at the
   moment, and wish to make only one change at a time.  -Karl */

# ifdef START_RSH_WITH_POPEN_RW

/* This is actually a crock -- it's OS/2-specific, for no one else
   uses it.  If I get time, I want to make piped_child and all the
   other stuff in os2/run.c work right.  In the meantime, this gets us
   up and running, and that's most important. */

void
start_rsh_server( cvsroot_t *root, struct buffer **to_server_p,
                  struct buffer **from_server_p )
{
    int pipes[2];
    int child_pid;

    /* If you're working through firewalls, you can set the
       CVS_RSH environment variable to a script which uses rsh to
       invoke another rsh on a proxy machine.  */
    char *cvs_rsh = getenv ("CVS_RSH");
    char *cvs_server = getenv ("CVS_SERVER");
    int i = 0;
    /* This needs to fit "rsh", "-b", "-l", "USER", "host",
       "cmd (w/ args)", and NULL.  We leave some room to grow. */
    char *rsh_argv[10];

    if (!cvs_rsh)
	/* People sometimes suggest or assume that this should default
	   to "remsh" on systems like HPUX in which that is the
	   system-supplied name for the rsh program.  However, that
	   causes various problems (keep in mind that systems such as
	   HPUX might have non-system-supplied versions of "rsh", like
	   a Kerberized one, which one might want to use).  If we
	   based the name on what is found in the PATH of the person
	   who runs configure, that would make it harder to
	   consistently produce the same result in the face of
	   different people producing binary distributions.  If we
	   based it on "remsh" always being the default for HPUX
	   (e.g. based on uname), that might be slightly better but
	   would require us to keep track of what the defaults are for
	   each system type, and probably would cope poorly if the
	   existence of remsh or rsh varies from OS version to OS
	   version.  Therefore, it seems best to have the default
	   remain "rsh", and tell HPUX users to specify remsh, for
	   example in CVS_RSH or other such mechanisms to be devised,
	   if that is what they want (the manual already tells them
	   that).  */
	cvs_rsh = RSH_DFLT;
    if (!cvs_server)
	cvs_server = "cvs";

    /* The command line starts out with rsh. */
    rsh_argv[i++] = cvs_rsh;

#   ifdef RSH_NEEDS_BINARY_FLAG
    /* "-b" for binary, under OS/2. */
    rsh_argv[i++] = "-b";
#   endif /* RSH_NEEDS_BINARY_FLAG */

    /* Then we strcat more things on the end one by one. */
    if (root->username != NULL)
    {
	rsh_argv[i++] = "-l";
	rsh_argv[i++] = root->username;
    }

    rsh_argv[i++] = root->hostname;
    rsh_argv[i++] = cvs_server;
    rsh_argv[i++] = "server";

    /* Mark the end of the arg list. */
    rsh_argv[i]   = (char *) NULL;

    if (trace)
    {
	fprintf (stderr, " -> Starting server: ");
	for (i = 0; rsh_argv[i]; i++)
	    fprintf (stderr, "%s ", rsh_argv[i]);
	putc ('\n', stderr);
    }

    /* Do the deed. */
    child_pid = popenRW (rsh_argv, pipes);
    if (child_pid < 0)
	error (1, errno, "cannot start server via rsh");

    /* Give caller the file descriptors in a form it can deal with. */
    make_bufs_from_fds (pipes[0], pipes[1], child_pid, to_server_p, from_server_p, 0);
}

# else /* ! START_RSH_WITH_POPEN_RW */

void
start_rsh_server (cvsroot_t *root, struct buffer **to_server_p, struct buffer **from_server_p)
{
    /* If you're working through firewalls, you can set the
       CVS_RSH environment variable to a script which uses rsh to
       invoke another rsh on a proxy machine.  */
    char *cvs_rsh = getenv ("CVS_RSH");
    char *cvs_server = getenv ("CVS_SERVER");
    char *command;
    int tofd, fromfd;
    int child_pid;

    if (!cvs_rsh)
	cvs_rsh = pick_rsh(root);	/* TONY CHANGE */
    if (!cvs_server)
	cvs_server = "cvs";

    /* Pass the command to rsh as a single string.  This shouldn't
       affect most rsh servers at all, and will pacify some buggy
       versions of rsh that grab switches out of the middle of the
       command (they're calling the GNU getopt routines incorrectly).  */
    command = xmalloc (strlen (cvs_server) + 8);

    /* If you are running a very old (Nov 3, 1994, before 1.5)
     * version of the server, you need to make sure that your .bashrc
     * on the server machine does not set CVSROOT to something
     * containing a colon (or better yet, upgrade the server).  */
    sprintf (command, "%s server", cvs_server);

    {
        char *argv[50];                 /* a wild guess */
	char **p = argv;

	*p++ = cvs_rsh;

/* TONY CHANGE */
	/* splice in any per user/host arguments pick up from .cvsrc */
	if (rsh_extra_argc > 0) {
	    int	i;

	    /* Note: the 10 is roughly empirically determined by looking
	     * at the source code here (6 is a more accurate number, but
	     * a few bytes on the stack never hurt anyone */
	    if (rsh_extra_argc + 10 > sizeof(argv)/sizeof(argv[0]))
	    {
	        fprintf (stderr, "Too many per user args.  Max is %d\n",
			  sizeof(argv)/sizeof(argv[0])-10);
		exit(1);
	    }
	    for (i = 0; i < rsh_extra_argc; i++)
	    {
		*p++ = rsh_extra_argv[i];
	    }
	}
/* TONY CHANGE */


	/* If the login names differ between client and server
	 * pass it on to rsh.
	 */
	if (root->username != NULL)
	{
	    *p++ = "-l";
	    *p++ = root->username;
	}

	*p++ = root->hostname;
	*p++ = command;
	*p++ = NULL;

	if (trace)
        {
	    int i;

            fprintf (stderr, " -> Starting server: ");
	    for (i = 0; argv[i]; i++)
	        fprintf (stderr, "%s ", argv[i]);
	    putc ('\n', stderr);
	}
	child_pid = piped_child (argv, &tofd, &fromfd);

	if (child_pid < 0)
	    error (1, errno, "cannot start server via rsh");
    }
    free (command);

    make_bufs_from_fds (tofd, fromfd, child_pid, to_server_p, from_server_p, 0);
}

# endif /* START_RSH_WITH_POPEN_RW */

#endif /* NO_EXT_METHOD */

#endif /* CLIENT_SUPPORT */
